/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.ws.mdp;

/**
 *
 * @author Karel Horak
 */
public class RussellGridWorld extends MDPProblem {
    
    private static final int CELL_WALL = 0;
    private static final int CELL_EMPTY = 1;
    private static final int CELL_REWARD = 2;
    private static final int CELL_PENALTY = 3;
    
    private int cells[][];
    private MDPState grid[][];
    private int height;
    private int width;
    
    private double rewardNegative = -1.0;
    private double rewardPositive =  1.0;
    
    private double probSuccess = 0.8;
    private double probRhs = 0.1;
    private double probLhs = 0.1;
    
    public RussellGridWorld(int height, int width) {
        grid = new MDPState[height][width];
        cells = new int[height][width];
        this.height = height;
        this.width = width;
    }
    
    public void setRewards(double negative, double positive) {
        this.rewardNegative = negative;
        this.rewardPositive = positive;
    }
    
    public void setProbabilities(double success, double rhs, double lhs) {
        this.probSuccess = success;
        this.probLhs = lhs;
        this.probRhs = rhs;
    }
    
    public void load(String world) {
        int strIdx = 0;
        for(int i = 0; i < height; i++) {
            for(int j = 0; j < width; j++, strIdx++) {
                switch(world.charAt(strIdx)) {
                    case '-':
                        cells[i][j] = CELL_EMPTY;
                        grid[i][j] = new MDPState("cell [" + i + "," + j + "]");
                        break;
                    case 'w':
                        cells[i][j] = CELL_WALL;
                        grid[i][j] = null;
                        break;
                    case 'r':
                        cells[i][j] = CELL_REWARD;
                        grid[i][j] = new MDPState("cell [" + i + "," + j + "] (reward)");
                        grid[i][j].setEvaluation(rewardPositive);
                        break;
                    case 'p':
                        cells[i][j] = CELL_PENALTY;
                        grid[i][j] = new MDPState("cell [" + i + "," + j + "] (penalty)");
                        grid[i][j].setEvaluation(rewardNegative);
                        break;
                    default:
                        throw new IllegalArgumentException("Invalid char " + world.charAt(strIdx) + " in the world definition.");
                }
            }
        }
        
        worldInit();
    }
    
    private void worldInit() {
        clearStates();
        
        for(int i = 0; i < height; i++) {
            for(int j = 0; j < width; j++) {
                if(grid[i][j] == null) continue;
                else if(cells[i][j] == CELL_REWARD || cells[i][j] == CELL_PENALTY) {
                    addState(grid[i][j]);
                    continue;
                }
                
                Vector v = new Vector(-1, 0);
                for(int k = 0; k < 4; k++) {
                    MDPAction a = new MDPAction("movement " + v);
                    
                    if(accessible(v.add(j, i))) {
                        a.addOutcome(grid[i+v.y][j+v.x], probSuccess);
                    } else {
                        a.addOutcome(grid[i][j], probSuccess);
                    }
                    
                    Vector vright = v.turnRight();
                    Vector vleft = v.turnLeft();
                    
                    if(accessible(vright.add(j, i))) {
                        a.addOutcome(grid[i+vright.y][j+vright.x], probRhs);
                    } else {
                        a.addOutcome(grid[i][j], probRhs);
                    }
                    
                    if(accessible(vleft.add(j, i))) {
                        a.addOutcome(grid[i+vleft.y][j+vleft.x], probLhs);
                    } else {
                        a.addOutcome(grid[i][j], probLhs);
                    }
                    
                    grid[i][j].addAction(a);
                    v = v.turnRight();
                    
                }
                
                addState(grid[i][j]);
            }
        }
    }
    
    private boolean accessible(Vector v) {
        if(v.x < 0 || v.x >= width || v.y < 0 || v.y >= height) return false;
        else return grid[v.y][v.x] != null;
    }
    
    public void display() {
        for(int i = 0; i < height; i++) {
            for(int j = 0; j < width; j++) {
                if(cells[i][j] == CELL_REWARD) {
                    System.out.print("reward  ");
                } else if(cells[i][j] == CELL_PENALTY) {
                    System.out.print("penalty ");
                } else if(cells[i][j] == CELL_WALL) {
                    System.out.print("wall    ");
                } else {
                    System.out.printf("%1.5f ", grid[i][j].getEvaluation());
                }
            }
            System.out.println("");
        }
    }
    
    public MDPState[][] getGrid() {
        return grid;
    }
    public int getHeight() {
        return height;
    }
    public int getWidth() {
        return width;
    }
    
    private class Vector {
        
        public final int x;
        public final int y;
        
        public Vector(int x, int y) {
            this.x = x;
            this.y = y;
        }
        
        public Vector add(int x, int y) {
            return new Vector(x + this.x, y + this.y);
        }
        
        public Vector turnRight() {
            return new Vector(-this.y, this.x);
        }
        public Vector turnLeft() {
            return new Vector(this.y, -this.x);
        }
        
        @Override
        public String toString() {
            return new StringBuilder(5)
                    .append('(')
                    .append(this.x)
                    .append(',')
                    .append(this.y)
                    .append(')')
                    .toString();
        }
        
    }
    
}
